"""A helper module to collect citations from a software package."""

import inspect

from citation_compass.docstring_utils import (
    extract_citation,
    extract_urls,
)


def get_object_full_name(obj):
    """Return the maximally qualified name of a thing.

    Parameters
    ----------
    obj : object
        The obj to get the name of.

    Returns
    -------
    str
        The fully qualified name of the thing.
    """
    # Try to determine the name of the "thing".
    if hasattr(obj, "__qualname__"):
        base_name = obj.__qualname__
    elif hasattr(obj, "__name__"):
        base_name = obj.__name__
    elif hasattr(obj, "__class__"):
        # If this is an object, use the class's name.
        base_name = obj.__class__.__qualname__
    else:
        raise ValueError(f"Could not determine the name of {obj}")

    # Get the string for the module (if we can find it).
    module = inspect.getmodule(obj)
    full_name = base_name if module is None else f"{module.__name__}.{base_name}"
    return full_name


class CitationEntry:
    """A (data)class to store information about a citation.

    Attributes
    ----------
    key : str
        The name of the module, function, or other aspect where the citation is needed.
    citation : str, optional
        The citation string.
    label : str, optional
        The (optional) user-defined label for the citation.
    urls : list of str
        A list of URLs extracted from the citation string.
    """

    def __init__(self, key, citation=None, label=None):
        self.key = key
        self.citation = citation
        self.label = label

        if citation is None:
            if label is not None and len(label) > 0:
                self.citation = label
            else:
                self.citation = "No citation provided."

        self.urls = extract_urls(self.citation)

    def __hash__(self):
        return hash(self.key)

    def __str__(self):
        return f"{self.key}: {self.citation}"

    def __repr__(self):
        return f"{self.key}:\n{self.citation}"

    @classmethod
    def from_object(cls, obj, label=None):
        """Create a CitationEntry from any object (including a function or method).

        Parameters
        ----------
        obj : object
            The object from which to create the citation.
        label : str, optional
            The (optional) user-defined label for the citation.

        Returns
        -------
        CitationEntry
            The citation entry.
        """
        # Try to parse a citation from the object's docstring (if there is one).
        if hasattr(obj, "__doc__"):
            docstring = obj.__doc__
        elif hasattr(obj, "__class__") and hasattr(obj.__class__, "__doc__"):
            docstring = obj.__class__.__doc__
        else:
            docstring = ""
        citation_text = extract_citation(docstring)
        if citation_text is None:
            citation_text = docstring

        full_name = get_object_full_name(obj)

        return cls(
            key=full_name,
            citation=citation_text,
            label=label,
        )


class CitationRegistry:
    """A class to store and manage citations for a software package.

    Attributes
    ----------
    all_entries : dict
        A dictionary mapping a annotation's key to its CitationEntry object.
    used_entries : set
        A set of keys for the "used" citations that have been used. This key matches
        the citation's key in the `all_entries` dictionary.
    used_trackers : dict
        A dictionary of additional user-defined trackers for used citations. Each
        tracker is a set of keys for the citations that have been used.
    """

    def __init__(self):
        self.all_entries = {}

        # We include a separate set of all used entries (instead of adding a "global"
        # entry to used_trackers) to avoid having to avoid having to do a dictionary
        # lookup for the "global" tracker. This adds some code complexity, but should
        # be faster in the common case.
        self.used_entries = set()
        self.used_trackers = {}

    def __len__(self):
        return len(self.all_entries)

    def __contains__(self, key):
        return key in self.all_entries

    def __getitem__(self, key):
        return self.all_entries[key]

    def add(self, key, citation):
        """Add a citation to the registry.

        Parameters
        ----------
        key : str
            The key for the citation.
        citation : CitationEntry, str, or object
            The citation entry to add.
        """
        if key not in self.all_entries:
            if isinstance(citation, CitationEntry):
                self.all_entries[key] = citation
            elif isinstance(citation, str):
                self.all_entries[key] = CitationEntry(key, citation)
            else:
                self.all_entries[key] = CitationEntry.from_object(citation)

    def mark_used(self, key):
        """Mark a citation as used.

        Parameters
        ----------
        key : str
            The key for the citation.
        """
        if key not in self.used_entries:
            # Check if we already have a global "used" citation for this key
            # and, if not, add it.
            self.used_entries.add(key)
        if len(self.used_trackers) > 0:
            # We avoid the loop if there are no trackers.  This is slightly
            # after than iterating over an empty dictionary.
            for tracker in self.used_trackers.values():
                if key not in tracker:
                    tracker.add(key)

    def num_trackers(self):
        """Return the number of custom trackers in use.

        Returns
        -------
        int
            The number of custom trackers.
        """
        # The number of custom trackers.
        return len(self.used_trackers)

    def start_used_tracker(self, name):
        """Add another tracker of used citations.

        Parameters
        ----------
        name : str
            The name of the tracker.
        """
        if name not in self.used_trackers:
            self.used_trackers[name] = set()
        else:
            raise KeyError(f"Tracker {name} already exists.")

    def stop_used_tracker(self, name):
        """Stop tracking used citations.

        Parameters
        ----------
        name : str
            The name of the tracker.

        Returns
        -------
        list of CitationEntry
            The list of used citations in the tracker.
        """
        if name in self.used_trackers:
            used = [self.all_entries[key] for key in self.used_trackers[name]]
            del self.used_trackers[name]
        else:
            raise KeyError(f"Tracker {name} does not exist.")
        return used

    def get_all_citations(self):
        """Return a list of all citations in the software package.

        Returns
        -------
        list of CitationEntry
            A list of all citations in the software package.
        """
        return list(self.all_entries.values())

    def get_used_citations(self, tracker_name=None):
        """Return a list of the used citations in the software package.

        Parameters
        ----------
        tracker_name : str, optional
            The name of the tracker to get the used citations for.
            If None, the global tracker is used.

        Returns
        -------
        list of CitationEntry
            A list of the citations used within the scope of the tracker.
        """
        if tracker_name is None:
            tracker = self.used_entries
        elif tracker_name in self.used_trackers:
            tracker = self.used_trackers[tracker_name]
        else:
            raise KeyError(f"Tracker {tracker_name} does not exist.")
        return [self.all_entries[key] for key in tracker]

    def reset_used_citations(self, tracker_name=None):
        """Reset the used citations for a tracker.

        Parameters
        ----------
        tracker_name : str, optional
            The name of the tracker to reset.
            If None, the global tracker is reset.
        """
        if tracker_name is None:
            self.used_entries = set()
        elif tracker_name in self.used_trackers:
            self.used_trackers[tracker_name] = set()
        else:
            raise KeyError(f"Tracker {tracker_name} does not exist.")


# ----------------------------------------------------------------------------
# The GLOBAL citation registry -----------------------------------------------
# ----------------------------------------------------------------------------

CITATION_COMPASS_REGISTRY = CitationRegistry()



import pytest

from citation_compass.citation_registry import (
    get_object_full_name,
    CitationEntry,
    CitationRegistry,
)


def example_function_1():
    """function_citation_1"""
    return 1


class ExampleClass:
    """A fake class for testing."""

    def __init__(self):
        pass

    def example_method(self):
        """A fake class method for testing."""
        return 0


def test_get_object_full_name():
    """Check that the full name is correctly generated."""
    assert get_object_full_name(example_function_1) == "test_citation_registry.example_function_1"
    assert (
        get_object_full_name(get_object_full_name)
        == "citation_compass.citation_registry.get_object_full_name"
    )
    assert get_object_full_name(ExampleClass) == "test_citation_registry.ExampleClass"
    assert (
        get_object_full_name(ExampleClass.example_method)
        == "test_citation_registry.ExampleClass.example_method"
    )

    obj = ExampleClass()
    assert get_object_full_name(obj) == "test_citation_registry.ExampleClass"
    assert get_object_full_name(obj.example_method) == "test_citation_registry.ExampleClass.example_method"


def test_citations_entry():
    """Check that we can create and query a citation entry."""
    entry = CitationEntry("key", "citation", "label")
    assert entry.key == "key"
    assert entry.citation == "citation"

    # If no citation is provided, the label is used.
    entry = CitationEntry("key", None, "label")
    assert entry.citation == "label"

    # We automatically extract URLs.
    entry = CitationEntry("key", "http://example.com")
    assert entry.urls == ["http://example.com"]

    # We can create a citation entry from an object.
    entry = CitationEntry.from_object(example_function_1)
    assert entry.key == "test_citation_registry.example_function_1"
    assert entry.citation == "function_citation_1"


def test_citations_registry():
    """Test that we can create and query a citation registry."""
    reg = CitationRegistry()
    assert len(reg) == 0

    # We can add a citation entry as a CitationEntry object.
    reg.add("key1", CitationEntry("key1", "citation1"))
    assert len(reg) == 1
    assert "key1" in reg
    assert reg["key1"].citation == "citation1"
    assert len(reg.get_all_citations()) == 1
    assert reg.get_all_citations()[0].key == "key1"

    # We can add a citation entry as a citation string.
    reg.add("key2", "citation2")
    assert len(reg) == 2
    assert "key2" in reg
    assert reg["key2"].citation == "citation2"
    assert len(reg.get_all_citations()) == 2

    # We can add a citation entry from an object.
    reg.add("key3", example_function_1)
    assert len(reg) == 3
    assert "key3" in reg
    assert reg["key3"].citation == "function_citation_1"
    assert len(reg.get_all_citations()) == 3

    # We can mark an entry as used.
    assert len(reg.get_used_citations()) == 0
    reg.mark_used("key1")
    assert len(reg.get_used_citations()) == 1
    assert reg.get_used_citations()[0].key == "key1"

    # We can add a custom tracker. Initially it has nothing seen.
    assert reg.num_trackers() == 0
    reg.start_used_tracker("tracker1")
    assert reg.num_trackers() == 1
    assert len(reg.get_used_citations()) == 1
    assert len(reg.get_used_citations("tracker1")) == 0

    # We can mark an entry as used in a custom tracker.
    reg.mark_used("key2")
    assert len(reg.get_used_citations()) == 2  # Global has 2
    assert len(reg.get_used_citations("tracker1")) == 1

    # We can reset the used citations for a tracker.
    reg.reset_used_citations("tracker1")
    assert len(reg.get_used_citations()) == 2  # Global still has 2
    assert len(reg.get_used_citations("tracker1")) == 0

    # We can re-mark an entry as used in a custom tracker.
    reg.mark_used("key2")
    assert len(reg.get_used_citations()) == 2  # Global still has 2
    assert len(reg.get_used_citations("tracker1")) == 1

    # We can't start another tracker with the same name.
    with pytest.raises(KeyError):
        reg.start_used_tracker("tracker1")

    # We can stop a tracker and get the used citations.
    used = reg.stop_used_tracker("tracker1")
    assert len(used) == 1
    assert used[0].key == "key2"
    assert len(reg.get_used_citations()) == 2  # Global still has 2

    assert reg.num_trackers() == 0
    with pytest.raises(KeyError):
        _ = reg.get_used_citations("tracker1")

    # We cannot stop or reset trackers that do not exist.
    with pytest.raises(KeyError):
        _ = reg.stop_used_tracker("tracker2")
    with pytest.raises(KeyError):
        reg.reset_used_citations("tracker2")



"""A helper module to collect citations from a software package."""

from functools import wraps
import logging
from os import urandom
import sys

from citation_compass.citation_registry import (
    get_object_full_name,
    CitationEntry,
    CITATION_COMPASS_REGISTRY,
)
from citation_compass.docstring_utils import extract_citation


def cite_module(name, citation=None):
    """Add a citation to a entire module.

    Parameters
    ----------
    name : str
        The name of the module.
    citation : str, optional
        The citation string. If None the code automatically tries to extract the citation text
        from the module's docstring.
    """
    if citation is None and name in sys.modules:
        module = sys.modules[name]
        if hasattr(module, "__doc__"):
            citation = extract_citation(module.__doc__)
            if citation is None or len(citation) == 0:
                citation = module.__doc__

    entry = CitationEntry(name, citation)
    CITATION_COMPASS_REGISTRY.add(name, entry)
    CITATION_COMPASS_REGISTRY.mark_used(name)


class CiteClass:
    """A super class for adding a citation to a class."""

    def __init__(self):
        pass

    def __init_subclass__(cls):
        # Add the class's full name
        full_name = get_object_full_name(cls)
        cls._citation_compass_name = full_name

        # Save the citation as ALL when it is first defined.
        if full_name not in CITATION_COMPASS_REGISTRY:
            CITATION_COMPASS_REGISTRY.add(full_name, CitationEntry.from_object(cls))
        else:
            logging.warning(f"Duplicated citation tag for class: {full_name}")

        # Wrap the constructor so the class is marked used when
        # the first object is instantiated.
        original_init = cls.__init__

        @wraps(original_init)
        def init_wrapper(*args, **kwargs):
            # Save the citation as USED when it is first called.
            CITATION_COMPASS_REGISTRY.mark_used(cls._citation_compass_name)
            return original_init(*args, **kwargs)

        cls.__init__ = init_wrapper


def cite_function(label=None, track_used=True):
    """A function wrapper for adding a citation to a function or
    class method.

    Parameters
    ----------
    label : str, optional
        The (optional) user-defined label for the citation.
    track_used : bool
        If True, the function will be marked as used when it is called.
        This adds a small amount of overhead to each function call.
        Default: True.

    Returns
    -------
    function
        The wrapped function or method.
    """
    # If the label is callable, there were no parentheses on the
    # dectorator and it passed in the function instead. So use None
    # as the label.
    use_label = label if not callable(label) else None

    def decorator(func):
        full_name = get_object_full_name(func)

        # Save the citation as ALL when it is first defined.
        if full_name not in CITATION_COMPASS_REGISTRY:
            citation = CitationEntry.from_object(func, label=use_label)
            CITATION_COMPASS_REGISTRY.add(full_name, citation)
        else:
            logging.warning(f"Duplicated citation tag for function: {full_name}")

        # Wrap the function so it is marked as USED when it is called.
        if track_used:

            @wraps(func)
            def fun_wrapper(*args, **kwargs):
                # Save the citation as USED when it is first called.
                CITATION_COMPASS_REGISTRY.mark_used(full_name)
                return func(*args, **kwargs)
        else:
            # We do not wrap the function, but just return the original function.
            fun_wrapper = func

            # We mark as used be default so the citation does not get dropped.
            CITATION_COMPASS_REGISTRY.mark_used(full_name)

        return fun_wrapper

    if callable(label):
        return decorator(label)
    return decorator


def cite_object(obj, label=None):
    """Add a citation for a specific object.

    Parameters
    ----------
    obj : object
        The object to add a citation for.
    label : str, optional
        The (optional) user-defined label for the citation.
    """
    full_name = get_object_full_name(obj)
    if full_name not in CITATION_COMPASS_REGISTRY:
        CITATION_COMPASS_REGISTRY.add(full_name, CitationEntry.from_object(obj, label=label))
    else:
        logging.warning(f"Duplicated citation tag for object: {full_name}")

    CITATION_COMPASS_REGISTRY.mark_used(full_name)


class CitationContext:
    """A context manager for tracking citations used within a block of code.

    Parameters
    ----------
    name : str, optional
        The name of the tracker. If None, a random name is generated.
    """

    def __init__(self, name=None):
        if name is None:
            # If no name is given, generate a random one.
            name = str(urandom(4))
        self.name = name

    def __enter__(self):
        CITATION_COMPASS_REGISTRY.start_used_tracker(self.name)
        return self

    def __exit__(self, *args):
        CITATION_COMPASS_REGISTRY.stop_used_tracker(self.name)

    def get_citations(self):
        """Return a list of all citations used within the context manager."""
        return [str(entry) for entry in CITATION_COMPASS_REGISTRY.get_used_citations(self.name)]


def get_all_citations():
    """Return a list of all citations in the software package.

    Returns
    -------
    citations : list of str
        A list of all citations in the software package.
    """
    citations = [str(entry) for entry in CITATION_COMPASS_REGISTRY.get_all_citations()]
    return citations


def get_used_citations(tracker_name=None):
    """Return a list of all citations in the software package.

    Parameters
    ----------
    tracker_name : str, optional
        The name of the tracker to get the used citations for.
        If None, the global tracker is used.

    Returns
    -------
    list of str
            A list of the citations used within the scope of the tracker.
    """
    citations = [str(entry) for entry in CITATION_COMPASS_REGISTRY.get_used_citations(tracker_name)]
    return citations


def find_in_citations(query, used_only=False, tracker_name=None):
    """Find a query string in the citation text. This is primarily used for
    testing, where a user might want to check if a citation is present.

    Parameters
    ----------
    query : str
        The query string to search for.
    used_only : bool, optional
        If True, only search in the used citations. If False, search in all citations.
    tracker_name : str, optional
        The name of the tracker to get the used citations for. Ignored if used_only is False.
        If None, and used_only=True, the global tracker is used.

    Returns
    -------
    matches : list of str
        A list of matching citation strings. This list is empty if no matches are found.
    """
    search_set = get_used_citations(tracker_name) if used_only else get_all_citations()
    matches = []
    for entry in search_set:
        if query in entry:
            matches.append(entry)
    return matches


def reset_used_citations(tracker_name=None):
    """Reset the list of used citations.

    Parameters
    ----------
    tracker_name : str, optional
        The name of the tracker to reset. If None, the global tracker is reset.
    """
    CITATION_COMPASS_REGISTRY.reset_used_citations(tracker_name)


def print_all_citations():
    """Print all citations in the software package in a user-friendly way."""
    for entry in CITATION_COMPASS_REGISTRY.get_all_citations():
        print(f"{entry.key}:\n{entry.citation}\n")


def print_used_citations(tracker_name=None):
    """Print the used citations in the software package in a user-friendly way.

    Parameters
    ----------
    tracker_name : str, optional
        The name of the tracker to reset. If None, the global tracker is reset.
    """
    for entry in CITATION_COMPASS_REGISTRY.get_used_citations(tracker_name):
        print(f"{entry.key}:\n{entry.citation}\n")
